/**
 * image2dsm version 1.0
 * Written by Steven "Sauce" Osman, sauce@playstation2-linux.com
 * Copyright (C) 2002, Steven Osman
 *
 * This program converts a jpeg to a .dsm file.
 *
 * Usage:
 * image2dsm <input filename> <image name> [<alpha value>]
 * <input filename> is the .jpeg file to convert
 * <image name> is the label to use for the data in the .dsm file
 * <alpha value> is the alpha value for the pixels, 0-255.  Defaults
 *  to 127.
 *
 * Output written to stdout, so you better redirect it!
 *
 * There may be issues about allowed image sizes.  Currently this program
 * converts images *exactly* as they are, without any rescaling.
 * Please make sure the image being converted is in a suitable
 * resolution.
 */
#define VERSION "1.0"

#include <stdlib.h>
#include <stdio.h>
#include <jpeglib.h>

/**
 * This structure will be able to work with different types of images
 * as this program is updated to support different file formats.
 */
typedef struct {
  char *pcFilename;

  enum {
    TYPE_JPEG,
    TYPE_UNKNOWN,
  } filetype;

  union {
    struct {
      struct jpeg_decompress_struct jpeg;
      struct jpeg_error_mgr jerr;
    } j;
  } source;

  FILE *pfInputFile;
} imageFile;


/**
 * Opens an image
 */
int openImage(imageFile *pImage, const char *pcImageFilename);

/**
 * Converts image
 */
void outputImage(imageFile *pImage,
		 const char *pcImageLabel,
		 unsigned int uiAlphaValue,
		 FILE *pfOutputFile);
/**
 * Closes image
 */
void closeImage(imageFile *pImage);

/**
 * Main routine.
 * Does nothing exciting - reads parameters, opens file, converts, closes file.
 */
int main(int iArgC, char **ppcArgV) {
  const char *pcFilename;
  const char *pcImageLabel;

  imageFile image;

  unsigned int uiAlpha=-1;

  if (iArgC<3) {
    fprintf(stderr,"Usage %s <image> <label> [<transparancy>]\n",ppcArgV[0]);
    exit(-1);
  }

  pcFilename=ppcArgV[1];
  pcImageLabel=ppcArgV[2];
  
  if (iArgC==4) {
    uiAlpha=atoi(ppcArgV[3]);
  }

  if (uiAlpha<0 || uiAlpha>255) {
    uiAlpha=127;
  }

  if (openImage(&image, pcFilename)) {
    fprintf(stderr,"Error opening source image\n");
    exit(-1);
  }

  outputImage(&image,pcImageLabel,uiAlpha,stdout);
  closeImage(&image);

  return 0;
}

/**
 * Opens up the image and reads in the header.
 * jpegImage points to the structure in which to load the image
 * pcImageFilename is the file name to load
 * returns 0 on success, non-zero on failure.
 */
int openImage(imageFile *pImage,
	      const char *pcImageFilename) {

  pImage->filetype=TYPE_UNKNOWN;
  pImage->pcFilename=0;
  pImage->pfInputFile=0;

  if ((pImage->pfInputFile=fopen(pcImageFilename, "rb"))==NULL) {
    fprintf(stderr, "Can't open %s\n",pcImageFilename);
    return -1;
  }

  pImage->pcFilename=malloc(strlen(pcImageFilename)+1);
  strcpy(pImage->pcFilename,pcImageFilename);

  pImage->filetype=TYPE_JPEG;

  pImage->source.j.jpeg.err = jpeg_std_error(&(pImage->source.j.jerr));
  jpeg_create_decompress(&(pImage->source.j.jpeg));
  

  jpeg_stdio_src(&(pImage->source.j.jpeg), pImage->pfInputFile);
  jpeg_read_header(&(pImage->source.j.jpeg), TRUE);

  return 0;
}


/**
 * Converts the image to DSM.
 * Currently only JPEG images are supported.
 * pImage is a poitner to the input image
 * pcImageLabel is the label the image should be named in the DSM file
 * uiAlphaValue is the alpha value (0-255) to use for the image pixels
 * pfOutputFile is the destination for the DSM image
 */
void outputImage(imageFile *pImage,
		 const char *pcImageLabel,
		 unsigned int uiAlphaValue,
		 FILE *pfOutputFile) {
  JSAMPARRAY buffer;
  JSAMPROW   row; 

  if (pImage->filetype!=TYPE_JPEG) {
    fprintf(stderr,"Unsupported file type can't be outputted.\n");
    exit(-1);
  }

  fprintf(pfOutputFile,
	  "/*\n/ * Produced by image2dsm v%s\n * Written by Steven \"Sauce\" Osman, sauce@playstation2-linux.com\n *\n * Source image %s\n * Dimensions %ix%i\n *\n */\n\n",
	  VERSION,
	  pImage->pcFilename,pImage->source.j.jpeg.image_width,
	  pImage->source.j.jpeg.image_height);

  fprintf(pfOutputFile,".macro iyx\ti1, i2\n.int\t\\i2, \\i1\n.endm\n\n.global %s\n.align 4\n\n.vu\n.data\n\n%s:\n",pcImageLabel,pcImageLabel);

  pImage->source.j.jpeg.out_color_components=3;
  pImage->source.j.jpeg.out_color_space=JCS_RGB;

  buffer = (*pImage->source.j.jpeg.mem->alloc_sarray) 
    ((j_common_ptr) &(pImage->source.j.jpeg), JPOOL_IMAGE, 
     pImage->source.j.jpeg.out_color_components * pImage->source.j.jpeg.image_width, 1); 

  jpeg_start_decompress(&(pImage->source.j.jpeg));
  
  uiAlphaValue<<=24;

  while (pImage->source.j.jpeg.output_scanline < pImage->source.j.jpeg.output_height) {
    int iColumn;
    unsigned uiPixel;

    jpeg_read_scanlines(&(pImage->source.j.jpeg), buffer, 1); 
    row = buffer[0];

    for (iColumn=0;iColumn<pImage->source.j.jpeg.output_width;iColumn++) {
      uiPixel=uiAlphaValue |
	(row[iColumn*3]) | 
	(row[iColumn*3+1] << 8) | 
	(row[iColumn*3+2] << 16);

      if (iColumn % 2) {
	fprintf(pfOutputFile,", 0x%08x\n",uiPixel);
      } else {
	fprintf(pfOutputFile,"iyx 0x%08x",uiPixel);
      }
    }

    fprintf(pfOutputFile,"\n");
  }

  jpeg_finish_decompress(&(pImage->source.j.jpeg));
}

/**
 * Closes an opened image.
 * pImage is a pointer to the image file to close.
 */
void closeImage(imageFile *pImage) {
  if (pImage->pcFilename) {
    free(pImage->pcFilename);
    pImage->pcFilename=0;
  }

  if (pImage->filetype==TYPE_JPEG) {
    jpeg_destroy_decompress(&(pImage->source.j.jpeg));
  }

  if (pImage->pfInputFile) {
    fclose(pImage->pfInputFile);
    pImage->pfInputFile=0;
  }
}
